package main

// this struct needs inteface

import (
    "strconv"
    "fmt"
    "math"
    )

type Agent struct {
    id              int
    age             int
    genome          *Genome
    cognition       Cognition
    x               int
    y               int
    fitness         float64
    fitnessCalcDone bool
    alive           bool
    kill            bool // default is killing if not said elsewat
    dialogsSuccess  int
    countSpeaker    int
    countHearer     int
    relations       int
}


func NewAgent(number, x, y int, alive, kill bool, globalLexicon *int) *Agent {
            cognitionNaminggame := NewCognitionNaminggame(globalLexicon)
            cognition := Cognition(cognitionNaminggame)
            genomeLength := GlobalOptions.geneScheme[len(GlobalOptions.geneScheme)-1]
            agent := Agent{id: number,
                           x: x,
                           y: y,
                           genome: NewGenome(genomeLength),
                           alive: alive,
                           kill: kill,
                           cognition: cognition}
            agent.Development()
            return &agent
}

func (a *Agent) Copy() *Agent {
    newAgent := *a
    newAgent.genome = a.genome.Copy()
    newAgent.cognition = a.cognition.Copy()
    return &newAgent
}

func (a *Agent) GetId() int {
    return a.id
}

func (a *Agent) Development() {
    a.cognition.Development(a.genome)
}

func (a *Agent) GetFitness() float64 {
    if (a.fitnessCalcDone == false) {
        a.CalcFitness()
    }
    return a.fitness
}

func (a *Agent) NewGenome(genome *Genome) {
    a.genome = genome
    a.Development()
}


func (a *Agent) CalcFitness() {
     a.fitness = 1
     // lipowska model
     //a.fitness = 1 - (math.Exp( -5 * float64(a.GetSumWeights() / 
     //                                  GlobalOptions.averageLexiconWeights)))

     // direct fitness on communication success
     // a.fitness = (float64(a.dialogsSuccess))/float64(a.GetDialogTotal()+1)
    
     // with a more unlimeted number of relations
      a.fitness = a.fitness * ( math.Exp( 0.02 * float64(a.relations) )  - 1)

     
      //a.fitness = a.fitness * (float64(a.relations)/float64(GlobalOptions.maxDegree*2))
      a.fitness = a.fitness * ( math.Exp( -0.05 * float64(a.GetAge()) ) )

    if(a.fitness < 0) {
        panic("Fitness can not be less then 0")
    }
    fmt.Sprintf("%v", math.Exp(1)) // does nothing, keeps imports warm
    a.fitnessCalcDone = true
}

func (a *Agent) Mutate(chance float64) {
    a.genome.Mutate()
    a.fitnessCalcDone = false
}

func (a *Agent) String() string  {
    return strconv.Itoa(a.id) +
            " - Fitness:" + FloatToString(a.GetFitness()) +
            " - Age: " + IntToString(a.age) + 
            " - Genome: " + a.genome.genome + "\n" +
            " - Total dialog: " + IntToString(a.GetDialogTotal()) + "\n" +
            " - Success dialog: " + IntToString(a.GetDialogSuccess()) + "\n" +
            " - Count as hearer: " + IntToString(a.GetCounterHearer()) + "\n" +
            " - Count as speaker: " + IntToString(a.GetCounterSpeaker()) + "\n" +
     //       " - Cost: " + FloatToString(a.GetCost()) + "\n" +
            " Memory: " + a.cognition.String()

}

func (a *Agent) Speak() Utterance {
    a.countSpeaker += 1
    return a.cognition.Speak()
}

func (a *Agent) Hear(utterance Utterance) (bool, float64) {
    // agent get utterance in a dialog
    a.countHearer += 1
    success, learningCost := a.cognition.Hear(utterance)
    if(success) {
        a.dialogsSuccess += 1
    }
    // involved in dialog, calc fitness on next pull
    a.fitnessCalcDone = false
    return success, learningCost
}


func (a *Agent) GetDialogTotal() int {
    return a.countSpeaker + a.countHearer
}

func (a *Agent) GetCounterHearer() int {
    return a.countHearer
}

func (a *Agent) GetCounterSpeaker() int {
    return a.countSpeaker
}

func (a *Agent) GetAge() int {
    return a.age
}

func (a *Agent) GetDialogSuccess() int {
    return a.dialogsSuccess
}

// in lack of a better name
func (a *Agent) DialogEnd(success bool, utterance Utterance, learningCost float64) {
    if(success) {
        a.dialogsSuccess += 1
    }
    // this  calc fitness on next pull
    a.fitnessCalcDone = false
    a.cognition.DialogEnd(success, utterance, learningCost)
}

func (a *Agent) GetSumWeights() float64 {
    return a.cognition.GetMemoryWeight()
}

func (a *Agent) GetLearnability() float64 {
    return a.cognition.GetLearnability()
}

func (a *Agent) GetHighestWeightedUtterance() Utterance {
    return a.cognition.GetHighestWeightedUtterance()
}

func (a *Agent) GiveFirstUtterance(utterance Utterance) {
    a.cognition.GiveFirstUtterance(utterance)
}

func (a *Agent) GetMemory() *map[Utterance]float64 {
    return a.cognition.GetMemory()
}

func (a *Agent) GetLearnabilityRate() float64 {
    panic("not in use")
}

func (a *Agent) IsAlive() bool {
    return a.alive
}

func (a *Agent) SpeakToForeignerChance() float64 {
    return a.cognition.SpeakToForeignerChance()
}

func (a *Agent) GetLearnChildProbability() float64 {
    return a.cognition.GetLearnChildProbability()
}

func (a *Agent) GetMemoryLimit() float64 {
    return a.cognition.GetMemoryLimit()
}

func (a *Agent) GetNumberOfKnownUtterances() int {
    return a.cognition.GetNumberOfKnownUtterances()
}

func (a *Agent) GetSpeechAbility() float64 {
    return a.cognition.GetSpeechAbility()
}
