package main

import (
    "fmt"
//    "math"
)

type CognitionNaminggame struct {
    memory           map[Utterance]float64
    globalLexicon    *int
    learnability     float64
    foreignChance    float64
    learnChildChance float64
    memoryLimit      float64
    speech           float64
}

func NewCognitionNaminggame(globalLexicon *int) *CognitionNaminggame {
    if(GlobalOptions.numberOfGenes < 4) {
        panic("CognitionNaminggame needs 4 gene")
    }
    memory := make(map[Utterance]float64)
    return &CognitionNaminggame{globalLexicon: globalLexicon,
                                memory: memory}
}

func (cn *CognitionNaminggame) Copy() Cognition {
    newCognition := *cn
    return &newCognition

}

func (cn *CognitionNaminggame) Development(genome *Genome)  {
    cn.learnability = genome.GetGene(GlobalOptions.geneScheme, 0)
    cn.foreignChance = genome.GetGene(GlobalOptions.geneScheme, 1)
    cn.learnChildChance = genome.GetGene(GlobalOptions.geneScheme, 2)
    cn.memoryLimit = genome.GetGene(GlobalOptions.geneScheme, 3)
    cn.speech = genome.GetGene(GlobalOptions.geneScheme, 4)
}

func (cn *CognitionNaminggame) Hear(utterance Utterance) (bool, float64) {
    if _,ok := cn.memory[utterance]; ok {
        // this utterance is known
        cn.memory[utterance] += cn.GetLearnability()
        return true, 0.0 // no costs assosiated with naming game, therefore 0
    }
    // never head of it, but adding it now
    // if memory at its limit, make place
    if(GlobalOptions.infiniteMemory == false) {
        hardLimit := GlobalOptions.populationSize / 2
        hardLimit = 20
        if(len(cn.memory) > int(cn.GetMemoryLimit()*float64(hardLimit))) {
            lowest := 10000.0
            var utteranceLowest Utterance
            for utterance, weight := range cn.memory {
                if weight < lowest {
                    utteranceLowest = utterance
                    lowest = weight
                }
            }
            delete(cn.memory, utteranceLowest)
        }
    }
    cn.memory[utterance] = cn.StartWeight()
    return false, 0.0
}

func (cn *CognitionNaminggame) Speak() Utterance {
    if(len(cn.memory) == 0 || (GlobalOptions.evolveSpeech && randFloat() > cn.speech)) {
        newUtterance := cn.MakeUtterance()
        cn.memory[newUtterance] = cn.StartWeight()
        return newUtterance
    }
    // put in own lib function 
    sum := 0.0
    for _, weight := range cn.memory {
        sum += weight
    }
    pick := float64(randInt(0, int(sum*1000))) / 1000.0
    sum = 0.0
    for utterance, weight := range cn.memory {
        sum += weight
        if sum >= pick {
            return utterance
        }
    }
    panic("Should never end up here")
}

func (cn *CognitionNaminggame) DialogEnd(success bool, utterance Utterance, learningCost float64) {
    if (success) {
        cn.memory[utterance] += cn.GetLearnability()
    } else {
        cn.memory[utterance] -= cn.GetLearnability()
        if cn.memory[utterance] < 0.0 {
            delete(cn.memory, utterance)
        }
    }
}


func (cn *CognitionNaminggame) MakeUtterance() Utterance {
    // A uniq word is representet as an uniq int.
    // Every int lower then cn.globalLexicon+1 are allready "invented"
    *cn.globalLexicon += 1
    return Utterance(IntToString(*cn.globalLexicon))
}


func (cn *CognitionNaminggame) GetLearnability() float64 {
    if GlobalOptions.fixedLearnability == true {
        return 1.0
    }
    return cn.learnability
}

func (cn *CognitionNaminggame) GetMemoryWeight() float64 {
    sum := 0.0
    for _, weight := range cn.memory {
        sum += weight
    }
    return sum
}

func (cn *CognitionNaminggame) GetHighestWeightedUtterance() Utterance {
    if (len(cn.memory) == 0) {
        return Utterance("0") // special utterance, means no utterance
    }
    weightBest := -1.0
    utteranceBest := *new(Utterance)
    for utterance, weight := range cn.memory {
        if weight > weightBest {
            weightBest = weight
            utteranceBest = utterance
        }
    }
    return utteranceBest
}

func (cn *CognitionNaminggame) GiveFirstUtterance(utterance Utterance) {
    cn.memory[utterance] = cn.StartWeight()
}

func (cn *CognitionNaminggame) String() string  {
    return fmt.Sprintf("%v", cn.memory)
}

func (cn *CognitionNaminggame) GetMemory() *map[Utterance]float64 {
    return &cn.memory
}

func (cn *CognitionNaminggame) StartWeight() float64 {
    if(GlobalOptions.startWeight < 0) {
        return cn.learnability
    }
    return GlobalOptions.startWeight
}

func (cn *CognitionNaminggame) SpeakToForeignerChance() float64 {
    return cn.foreignChance
}

func (cn *CognitionNaminggame) GetLearnChildProbability() float64 {
    return cn.learnChildChance
}

func (cn *CognitionNaminggame) GetMemoryLimit() float64 {
    return cn.memoryLimit
}

func (cn *CognitionNaminggame) GetNumberOfKnownUtterances() int {
    return len(cn.memory)
}

func (cn *CognitionNaminggame) GetSpeechAbility() float64 {
    return cn.speech
}
