package main

import (
    "fmt"
    "math/rand"
    "time"
    "flag"
    "os"
    "runtime/pprof"
    "log"
)

var maxGenerations int
var populationSize int
var geneLength int
var numberOfGenes int
var numberOfSimulations int
var crossoverPoints int
var startWeight float64
var statsSnapshot int
var dialogPrAgentPrGeneration int
var mutationChance float64
var turnamentSlection float64
var turnamentWinners float64
var httpserver bool
var notation string
var structure = flag.String("structure", "graph", "Type of structure ['graph', 'lattice']")
var game = flag.String("game", "naminggame", "Type of language game ['naminggame'")
var maxDegree = flag.Int("degree", 5, "Max degree on nodes if using graph as structure type")
var fixedLearnability = flag.Bool("fixedLearnability", false, "Set fixed learnability (i.e. disables it)")
var evolveSpeech = flag.Bool("evolveSpeech", false, "Add a gene for speech ability")
var infiniteMemory = flag.Bool("infiniteMemory", false, "Do not restrict inventory size")
var nographs bool
var cpuprofile = flag.String("cpuprofile", "", "write cpu profile to file")

var GlobalOptions Options



func init() {
    flag.IntVar(&populationSize, "p", 300, "Size of population")
    flag.IntVar(&geneLength, "bits", 8, "Number of bits pr gene")
    flag.IntVar(&numberOfGenes, "genes", 5, "Number of genes")
    flag.IntVar(&dialogPrAgentPrGeneration, "d", 5, "Number of dialogs pr agents pr generation")
    flag.IntVar(&statsSnapshot, "snap", 50, "Take heatmap snapshot pr x genreation")
    flag.IntVar(&maxGenerations, "i", 100, "Max generations")
    flag.IntVar(&crossoverPoints, "c", 1, "Number of crossover points")
    flag.StringVar(&notation, "notation", "", "Remark or notation to simulation")
    flag.Float64Var(&startWeight, "startweight", 1, "Initial weight on invented " +
                                "utteranutterancen, negative if learnabilit should be used")
    flag.IntVar(&numberOfSimulations, "s", 1, "Number of simulations (and average results)")
    flag.Float64Var(&mutationChance, "m", 0.01, "Mutation chance")
    flag.Float64Var(&turnamentSlection, "t", 0.8, "Proportion of agent in turnament selection")
    flag.Float64Var(&turnamentWinners, "w", 0.6, "Number of best agent in selection")
    flag.BoolVar(&httpserver, "http", true, "For httpserver for stat files")
    flag.BoolVar(&nographs, "nographs", false, "Turn off graphing, smart with profiling")

}

func main() {
    flag.Parse()

    // we need some randomness
    rand.Seed( time.Now().UTC().UnixNano())

    // if parameter given; we profile the simulation
    if *cpuprofile != "" {
        f, err := os.Create(*cpuprofile)
        if err != nil {
            log.Fatal(err)
        }
        pprof.StartCPUProfile(f)
        defer pprof.StopCPUProfile()
    }

    if turnamentSlection < turnamentWinners {
        fmt.Println("Error, turnament selection must be bigger or equal " +
                    "to turnament winners.")
    }


    // start point
    start := time.Now()
    Log("Evolang.go launched at " + time.Now().String())

    // load options into an struct used systemwide
    geneScheme := MakeGeneScheme(geneLength, numberOfGenes)
    options := Options{
        maxGenerations: maxGenerations,
        populationSize: populationSize,
        mutationChance: mutationChance,
        turnamentSlection: turnamentSlection,
        turnamentWinners: turnamentWinners,
        statsSnapshot: statsSnapshot,
        numberOfSimulations: numberOfSimulations,
        geneScheme: geneScheme,
        crossoverPoints: crossoverPoints,
        nographs: nographs,
        notation: notation,
        startWeight: startWeight,
        dialogPrAgentPrGeneration: dialogPrAgentPrGeneration,
        geneLength: geneLength,
        structure: *structure,
        numberOfGenes: numberOfGenes,
        game: *game,
        maxDegree: *maxDegree,
        fixedLearnability: *fixedLearnability,
        evolveSpeech: *evolveSpeech,
        infiniteMemory: *infiniteMemory,
        averageLexiconWeights: startWeight,
        crossoverOnGenes: true}

    GlobalOptions = options // global scope options struct

    fmt.Println(options)

    tester := NewTester()
    tester.Run()

    // Run the simulation
    solver := NewSolver()
    solver.Run()

    // Simulation is done, print runtime
    Log("Evolang.go ends - totalt time: " + time.Since(start).String() + "\n\n\n")
    fmt.Print("Total time: ")
    fmt.Println(time.Since(start))

    // if http flag set, present stats trough web server
    if(httpserver) {
        httpserver := NewHttpServer()
        httpserver.Run()
    }

}
