package main

import ( 
 //   "strconv"
    //"fmt"
    "sort"
    "strings"
)

type Genome struct {
    genome string
}

func NewGenome(length int) *Genome {
    bitString := randomBitString(length)
    return &Genome{genome: bitString}
}

func (g *Genome) Copy() *Genome {
    newGenome := *g
    return &newGenome
}

func (g *Genome) GetRepresentation() string {
    return string(g.genome)
}

func (g *Genome) GetGene(geneScheme []int, geneNumber int) float64 {
    // geneScheme could be [4, 9, 23] where there would be a total of 3 genes
    // ranged from bit 0-4, 5-9, 10-23
    end := geneScheme[geneNumber]
    var start int
    if geneNumber == 0 {
        start = 0
    } else {
        start = geneScheme[geneNumber-1]
    }
    bitString := g.genome[start:end]
    return FloatGivenMaxIntSize(bitString)
}

func (g *Genome) Mutate()  {
    g.genome = FlipBitsByChance(g.genome, GlobalOptions.mutationChance)
}

func (g *Genome) Crossover(genome2 *Genome) {
    // changes g gene, uses points and strategies from GlobalOptions
    genome1 := g
    genomeLength := len(genome1.genome)
    newGenome1 := make([]string, GlobalOptions.crossoverPoints+1)
    newGenome2 := make([]string, GlobalOptions.crossoverPoints+1)
    // store points in points
    points := make([]int, GlobalOptions.crossoverPoints)
    if GlobalOptions.crossoverOnGenes == true {
        // crossover exactly on genes
        for p := 0; p < GlobalOptions.crossoverPoints; p++ {
            points[p] = randInt(1, GlobalOptions.numberOfGenes-1) * GlobalOptions.geneLength
        }
    } else {
        // crossover on random points
        for p := 0; p < GlobalOptions.crossoverPoints; p++ {
            points[p] = randInt(0, genomeLength-1)
        }
    }
    // crossover
    sort.Ints(points)
    startpoint := 0
    for i, p := range points {
        if (i % 2 == 0) {
            newGenome1[i] = genome1.genome[startpoint:p]
            newGenome2[i] = genome2.genome[startpoint:p]
        } else {
            newGenome2[i] = genome1.genome[startpoint:p]
            newGenome1[i] = genome2.genome[startpoint:p]
        }
        startpoint = p
    }
    newGenome1[len(points)] = genome1.genome[startpoint:]
    newGenome2[len(points)] = genome2.genome[startpoint:]
    if randInt(0,1) == 1 {
        g.genome = strings.Join(newGenome1, "")
    } else {
        g.genome = strings.Join(newGenome2, "")
    }
}


