package main

import (
        "fmt"
//        "strconv"
//        "runtime"
        "github.com/sbinet/go-gnuplot"

)

type Graph struct {
    plotter    *gnuplot.Plotter
}

func NewGraph(title string, persist bool, debug bool) *Graph {
    plotter,err := gnuplot.NewPlotter("", persist, debug)
    if err != nil {
        err_string := fmt.Sprintf("** err: %v\n", err)
        panic(err_string)
    }
    graph := Graph{plotter: plotter}
    graph.SetTitle(title)
    return &graph
}

func (g *Graph) AddX(array []float64, name string) {
    points := 1000
    if (len(array) > points) {
        modulu := len(array) / points
        newArray := make([]float64, points)
        for i := range newArray{
            newArray[i] = array[i*modulu]
        }
        g.plotter.PlotX(newArray, name)
    } else {
        //fmt.Println(array)
        g.plotter.PlotX(array, name)
    }
}

func (g *Graph) Save(filename string) {
    g.plotter.CheckedCmd("set terminal png")
    g.plotter.CheckedCmd("set output '"+filename+".png'")
    g.plotter.CheckedCmd("replot")
    defer g.plotter.Close()
}

func (g *Graph) SetTitle(title string) {
    g.plotter.CheckedCmd("set title '"+title+"'")
    g.plotter.SetStyle("lines")
}

func (g *Graph) SetYRange(lower, upper int) {
    lowerString := IntToString(lower) 
    upperString := IntToString(upper) 
    g.plotter.CheckedCmd("set yrange ["+lowerString+":"+upperString+"]")
}


func (g *Graph) Replot() {
    g.plotter.CheckedCmd("replot")
}

func (g *Graph) RemoveLegend() {
    g.plotter.CheckedCmd("set nokey")
}


func (g *Graph) SetLabels(x, y string) {
    g.plotter.SetXLabel(x)  
    g.plotter.SetYLabel(y)  
}
