package main

import (
        "fmt"
        "bytes"
//        "strconv"
//        "runtime"
        "github.com/sbinet/go-gnuplot"

)

type HeatMap struct {
    plotter    *gnuplot.Plotter
}

func NewHeatMap(title string, persist bool, debug bool) *HeatMap {
    plotter,err := gnuplot.NewPlotter("", persist, debug)
    if err != nil {
        err_string := fmt.Sprintf("** err: %v\n", err)
        panic(err_string)
    }
    //heatMap.SetTitle(title)
    plotter.CheckedCmd("set title '"+title+"' ")
    heatMap := HeatMap{plotter: plotter}
    return &heatMap
}

func (g *HeatMap) UpdateMap(matrix [][]float64) {
    g.plotter.ResetPlot()
    //g.plotter.CheckedCmd("set cbrange [0:1]\n")
    var buffer bytes.Buffer
    for x := range matrix {
        for y := range matrix[x] {
            //if (!matrix[x][y].Alive) { buffer.WriteString("0 "); continue }
            buffer.WriteString(FloatToString(matrix[x][y]))
            buffer.WriteString(" ")
        }
        buffer.WriteString("\n")
    }
    buffer.WriteString("e\ne\n")
    g.plotter.CheckedCmd(buffer.String())

}

func (g *HeatMap) Save(filename string) {
    g.plotter.CheckedCmd("set terminal pngcairo transparent enhanced font 'arial,10' fontscale 1.0")
    g.plotter.CheckedCmd("set output '"+filename+".png'")
}

func (g *HeatMap) SetRanges(x,y,c float64, lable string) {
    g.plotter.CheckedCmd("unset key")
    g.plotter.CheckedCmd("set view map")
    g.plotter.CheckedCmd("set xtics border in scale 0,0 mirror norotate  offset character 0, 0, 0 autojustify")
    g.plotter.CheckedCmd("set ytics border in scale 0,0 mirror norotate  offset character 0, 0, 0 autojustify")
    g.plotter.CheckedCmd("set ztics border in scale 0,0 nomirror norotate  offset character 0, 0, 0 autojustify")
    g.plotter.CheckedCmd("set nocbtics")
    g.plotter.CheckedCmd("set rtics axis in scale 0,0 nomirror norotate  offset character 0, 0, 0 autojustify")
    xrange := FloatToString(x - 0.5)
    yrange := FloatToString(y - 0.5)
    g.plotter.CheckedCmd("set xrange [ -0.500000 : "+xrange+" ] noreverse nowriteback")
    g.plotter.CheckedCmd("set yrange [ -0.500000 : "+yrange+" ] noreverse nowriteback")
    g.plotter.CheckedCmd("set cblabel '"+lable+"' ")
    if(c > 0) {
        crange := FloatToString(c)
        g.plotter.CheckedCmd("set cbrange [ 0.00000 : "+crange+" ] noreverse nowriteback")
    }
    //p.CheckedCmd("set palette rgbformulae -7, 2, -7")
    g.plotter.CheckedCmd("splot '-' matrix with image")
}

func (g *HeatMap) SetTitle(title string) {
    g.plotter.CheckedCmd("set title 'Heat map' ")
}

func (g *HeatMap) ResetPlot() {
    g.plotter.ResetPlot()
}

func (g *HeatMap) SetLabels(x, y string) {
    g.plotter.SetXLabel(x)
    g.plotter.SetYLabel(y)
}

func (g *HeatMap) Close() {
    g.plotter.Close()
}
