package main

import (
        //"strconv"
        "fmt"
)

type AgentPair struct {
    a, b *Agent
}

type PopulationGraph struct {
    agents map[int]*Agent
    agentsAlive []*Agent
    individualCounter int
    globalLexicon *int
    relations map[AgentPair]float64
    
}

func NewPopulationGraph() *PopulationGraph {
    globalLexicon := 0
    return &PopulationGraph{
        globalLexicon:          &globalLexicon,
        agents:                 make(map[int]*Agent),
        agentsAlive:            make([]*Agent, GlobalOptions.populationSize),
        relations:              make(map[AgentPair]float64),
        individualCounter:      0}
}

func (pl *PopulationGraph) Populate() {
    fmt.Println("Populationg graph with", GlobalOptions.populationSize, "agents")
    for i := 0; i < GlobalOptions.populationSize; i ++  {
        agent := NewAgent(pl.individualCounter, 0, 0, true, true,
pl.globalLexicon) // agents coordinates are not used in social netwrok
        pl.agents[pl.individualCounter] = agent
        pl.agentsAlive[i] = agent
        pl.individualCounter+=1
    }
}

func (pl *PopulationGraph) Size() int {
    return len(pl.agentsAlive)
}

func (pl *PopulationGraph) Age() {
    for i, _ := range pl.agentsAlive { 
        pl.agentsAlive[i].age+=1
    }
}

func (pl *PopulationGraph) GetGenerationStats() GenerationStats {
    agentList := pl.MakeListOfAgents()
    stats := NewStats()
    generationStats := stats.GenerateStatsFromGeneration(agentList)
    generationStats.degreeAvg = pl.GetDegreeAvg()
    generationStats.clusteringCoefAvg = pl.GetClusterCoefficientAvg()
    return generationStats
}

func (pl *PopulationGraph) PrintRepresentation() {
//    for debuging
//    g := gographviz.NewGraph()
//    g.SetName("G")
//    g.SetDir(false)
//    for _, agent := range pl.agentsAlive {
//        //_ = pl.GetNeighbours(agent, false) // remove dead neighbours if exists
//        a := gographviz.NewAttrs()
//        a.Add("group", IntToString(randInt(0,3)))
//        g.AddNode("G", IntToString(agent.GetId()), a)
//    }
//    for agentPair, _ := range pl.relations {
//        a := gographviz.NewAttrs()
//        //a.Add("value", 0.1)
//        g.AddEdge(IntToString(agentPair), "", IntToString(neighbour.GetId()), "", false, a)
//    }
//    s := g.String()
//    ioutil.WriteFile("test.dot", []byte(s), 0666)
}

func (pl *PopulationGraph) Interact() {
    maxIterations := pl.Size() * GlobalOptions.dialogPrAgentPrGeneration
    for i := 0; i < maxIterations; i++ {
        agent1 := pl.GetRandomAgentProportioanlToDegree()
        chanceOfNewNeighbour := agent1.SpeakToForeignerChance()
        var agent2 *Agent

        for true {
            neighbours := pl.GetNeighbours(agent1, true)

            // if max degree is reached, choose one neighbour
            if(len(neighbours) >= GlobalOptions.maxDegree) {
                agent2 = neighbours[randInt(0, len(neighbours)-1)]

            // if no relations, make a random one
            } else if(len(neighbours) == 0) {
                agent2 = pl.GetRandomAgent() 

            // agent have something in between, find new agent out in graph
            // should have been done recurivly(!)
            } else {
                agent2 = neighbours[randInt(0, len(neighbours)-1)]
                randomFloat := randFloat()
                if(randomFloat < chanceOfNewNeighbour) {
                    neighbours = pl.GetNeighbours(agent2, true)
                    if(len(neighbours) > 1) { 
                        for true {
                            agent2 = neighbours[randInt(0, len(neighbours)-1)]
                            if agent1 != agent2 {
                                break
                            }
                        }
                        // one more time
                        randomFloat := randFloat()
                        if(randomFloat < chanceOfNewNeighbour) {
                            neighbours = pl.GetNeighbours(agent2, true)
                            if(len(neighbours) > 1) { 
                                for true {
                                    agent2 = neighbours[randInt(0, len(neighbours)-1)]
                                    if agent1 != agent2 {
                                        break
                                    }
                                }
                                // last time
                                randomFloat := randFloat()
                                if(randomFloat < chanceOfNewNeighbour) {
                                    neighbours = pl.GetNeighbours(agent2, true)
                                    if(len(neighbours) > 1) { 
                                        for true {
                                            agent2 = neighbours[randInt(0, len(neighbours)-1)]
                                            if agent1 != agent2 {
                                                break
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
            // ensure we have not choosen the same agent
            if agent1 != agent2 {
                break
            }
        }

        // do the actuall dialog
        pl.Dialog(agent1, agent2)
    }
}

func (pl *PopulationGraph) Dialog(agent1, agent2 *Agent) {

    // Random agents speaks
    utterance := agent1.Speak();
    // Random neighbour hears
    success, learningCost := agent2.Hear(utterance)
    // update agent1s memory 
    // in lack of a better name
    agent1.DialogEnd(success, utterance, learningCost)
    //success = true // for testing the model
    agentPair := GetAgentPair(agent1, agent2)
    if(success) {
        if _,ok := pl.relations[agentPair]; !ok {
            agent1.relations+=1
            agent2.relations+=1
            pl.relations[agentPair] = 1 // float64(GlobalOptions.generation)
        } else {
              pl.relations[agentPair] += 1
          }
    } else {
        if _,ok := pl.relations[agentPair]; ok {
            pl.relations[agentPair] -= 0.5
            if pl.relations[agentPair] <= 0 {
                agent1.relations-=1
                agent2.relations-=1
                delete(pl.relations, agentPair)
            }
        }
    }
}

func (pl *PopulationGraph) KillWeak() {
    turnamentSlection := int(GlobalOptions.turnamentSlection*float64(pl.Size()))
    turnamentWinners := int(GlobalOptions.turnamentWinners*float64(pl.Size()))
    // shuffel the list
    agents := ShuffelAgentArray(pl.agentsAlive)
    // get random k agents to turnament 
    turnamentAgents := agents[0:turnamentSlection]
    // sort them after fitness and choose the w best 
    turnamentAgents = SortAgentsByFitness(turnamentAgents)
    for i := 0; i < turnamentWinners; i++ {
        turnamentAgents[i].kill = false // dont kill the choosen ones
    }
    newAgentsAliveArray := make([]*Agent, turnamentWinners)
    counter := 0
    for a := range agents {
        if agents[a].kill == true {
            agents[a].alive = false // flagged as dead, 
        } else {
            // all the alive ones
            newAgentsAliveArray[counter] = agents[a]
            counter+=1
        }
        agents[a].kill = true // set everyone as candidates for next generation killing
    }
    pl.agentsAlive = newAgentsAliveArray
    for agentPair, _ := range pl.relations {
        if (agentPair.a.alive == false || agentPair.b.alive == false) {
            delete(pl.relations, agentPair)
        }
    }
    //pl.PrintRepresentation()
}

func (pl *PopulationGraph) Breed() {
    numberOfNewAgents := GlobalOptions.populationSize - len(pl.agentsAlive)
    agents := SortAgentsByFitness(pl.agentsAlive)
    for i := 0; i < numberOfNewAgents; i++ {
        agent1 := agents[i]
        var agent2 *Agent
        neighbours := pl.GetNeighbours(agent1, true)
        if(len(neighbours) > 0) {
            agent2 = neighbours[randInt(0, len(neighbours)-1)]
        } else {
            // ensure not selection the same agent
            for true {
                agent2 = pl.GetRandomAgent() 
                if agent1 != agent2 {
                    break
                }
            }
        }
        parent1 := agent1
        parent2 := agent2

        genome := parent1.genome.Copy()
        genome.Crossover(parent2.genome)
        // mutate with a given chance
        if randFloat() < GlobalOptions.mutationChance {
            genome.Mutate()
        }
        agent := NewAgent(pl.individualCounter, 0, 0, true, true, pl.globalLexicon)
        agent.NewGenome(genome)

        pl.agentsAlive = append(pl.agentsAlive, agent)
        pl.individualCounter+=1

        if(parent1.GetLearnChildProbability() > randFloat()) {
            pl.Dialog(parent1, agent)
        }
        if(parent2.GetLearnChildProbability() > randFloat()) {
            pl.Dialog(parent2, agent)
        }

    }
}

func (pl *PopulationGraph) StatsSnapshot() (*Snapshot) {
    agentsAliveCopy := make([]*Agent, len(pl.agentsAlive))
    for i := range agentsAliveCopy {
        agentsAliveCopy[i] = pl.agentsAlive[i].Copy()
    }
    relationsCopy := make(map[AgentPair]float64) 
    for agentPair, weight := range pl.relations {
        agentPairCopy := AgentPair{agentPair.a.Copy(), agentPair.b.Copy()}
        weightCopy := weight
        relationsCopy[agentPairCopy] = weightCopy
    }
    snapshot := NewSnapshot("graph")
    snapshot.AddGraphData(agentsAliveCopy, relationsCopy)
    return snapshot
}

// help functions

func (pl *PopulationGraph) GetRandomAgent() (*Agent) {
    list := pl.MakeListOfAgents()
    return list[randInt(0, len(list)-1)]
}

func (pl *PopulationGraph) GetRandomAgentProportioanlToDegree() (*Agent) {
    sum := 0.0
    for _, agent := range pl.agentsAlive {
        sum += float64(agent.relations+1)
    }
    decision := randFloat()*sum
    sum = 0.0
    for _, agent := range pl.agentsAlive {
        sum += float64(agent.relations+1)
        if sum >= decision {
            return agent
        }
    }
    return nil
}

func (pl *PopulationGraph) MakeListOfAgents() ([]*Agent) {
    return pl.agentsAlive // copy
}

func (pl *PopulationGraph) GetNeighbours(agent *Agent, returnOnlyAlive bool) ([]*Agent) {
    // finding neighbours
    neighbours := make([]*Agent, 0)
    for agentPair, _ := range pl.relations {
        if(agent == agentPair.a) {
            neighbours = append(neighbours, agentPair.b)
        }
        if(agent == agentPair.b) {
            neighbours = append(neighbours, agentPair.a)
        }
    }
    return neighbours
}

func (pl PopulationGraph) GetListOfAgentsWithRelations() map[*Agent][]*Agent {
    agentsRelations := make(map[*Agent][]*Agent)
    for agentPair, _ := range pl.relations {
        agentsRelations[agentPair.a] = append(agentsRelations[agentPair.a],
            agentPair.b)
        agentsRelations[agentPair.b] = append(agentsRelations[agentPair.b],
            agentPair.a)
    }
    return agentsRelations

}

func (pl *PopulationGraph) Consensus() bool {
    utterance := pl.agentsAlive[0].GetHighestWeightedUtterance()
    var newUtterance Utterance
    for _, agent := range pl.agentsAlive { 
        newUtterance = agent.GetHighestWeightedUtterance()
        if utterance != newUtterance {
            return false
        }
    }
    return true
}

func (pl PopulationGraph) GetDegreeAvg() float64 {
    // find the graphs average degree
    degreeAvg := 0.0
    for _, neighbors := range pl.GetListOfAgentsWithRelations() {
        degreeAvg += float64(len(neighbors))
    }
    return degreeAvg/float64(pl.Size())
}

func (pl PopulationGraph) GetClusterCoefficientAvg() float64 {
    clusterCoefficientAvg := 0.0
    agentsRelations := pl.GetListOfAgentsWithRelations() 
    for _, neighbors := range agentsRelations {
        numberOfNeighbors := len(neighbors)
        if numberOfNeighbors > 1 {
            edges := 0.0
            for _,i := range neighbors {
                for _,j := range neighbors {
                    if j == i {
                        break
                    }
                    if _,ok := pl.relations[GetAgentPair(i,j)]; ok {
                        edges += 1
                    }
                }
            }
            potensialEdges := numberOfNeighbors * (numberOfNeighbors - 1) / 2
            clusterCoefficientAvg += edges / float64(potensialEdges)
        }
    }
    if clusterCoefficientAvg == 0 {
        return 0
    }
    return clusterCoefficientAvg/float64(pl.Size())
}

func GetAgentPair(a, b *Agent) AgentPair {
    if a == b {
        panic("AgentPair cannot be the same")
    }
    if a.GetId() < b.GetId() {
        return AgentPair{a, b}
    }
    return AgentPair{b, a}
}

func InAgentPair(agent *Agent, agentPair AgentPair) bool {
    if agent == agentPair.a || agent == agentPair.b {
        return true
    }
    return false
}

