package main

import (
        "fmt"
        "strconv"
//        "runtime"
)

type Solver struct {
}

func NewSolver() *Solver {
    return &Solver{}
}

func (s *Solver) Run() {
    Log("Starting simulation")
    stats := NewStats()

    for simulation := 0; simulation < GlobalOptions.numberOfSimulations; simulation++ {
        Log("Simulation: " + strconv.Itoa(simulation))
        var population Population
        switch GlobalOptions.structure {
            case "lattice": population = Population(NewPopulationLattice())
            case "graph": population = Population(NewPopulationGraph())
        }
        population.Populate()
        for generation := 0; generation < GlobalOptions.maxGenerations; generation++ {
            GlobalOptions.generation = generation
            percentDone := float64((simulation+1) * (generation+1)) / 
                           float64(GlobalOptions.numberOfSimulations * GlobalOptions.maxGenerations) * 100
            //Log("Generation: " + strconv.Itoa(generation))
            fmt.Printf("%f \n", percentDone)

            // Interaction
            population.Age()
            population.Interact()

            // Generate stats
            generationStats := population.GetGenerationStats()
            stats.AddGenerationStats(simulation, generation, generationStats)

            // take snapshot of population
            if (generation % GlobalOptions.statsSnapshot == 0) {
                snapshot := population.StatsSnapshot()
                snapshot.AddGeneration(generation)
                stats.AddSnapshot(simulation, snapshot)
            }

            // kill and breed
            population.KillWeak()
            population.Breed()
         }
//         population.PrintRepresentation()
         stats.PlotFitnessHeatMapForSimulation(simulation)
         stats.PlotHighestWeightedUtteranceHeatMapForSimulation(simulation)
         stats.PlotLearnabilityRateHeatMapForSimulation(simulation)
    }
    stats.PrintSummary()
    if(GlobalOptions.nographs == false) {
        stats.WriteFitnessStatsToFiles()
        stats.MakeSimulationSummaryGraphs()
        stats.MakeGenomeSummaryGraphs()
        stats.MakeFitnessGraphs()
        stats.MakeLearnChildGraphs()
        stats.MakeMemoryLimitGraphs()
        //stats.MakeSumWeightsGraphs()
        //stats.MakeLearningBiasesGraphs()
        stats.MakeLearnabilityGraphs()
        //stats.MakeCostGraphs()
        stats.MakeGraphGraphs()
        stats.MakeExtrovertGraphs()
        stats.MakeSpeechGraphs()
        stats.MakeAgeGraphs()
        stats.MakeUtteranceGraphs()
        stats.MakeDialogGraphs()
        stats.CopyStatsFolderToCurrent()
    }
}


