package main

import (
        "fmt"
        "code.google.com/p/gographviz"
        "io/ioutil"
        "time"
        "strconv"
        "os/exec"
        "math"
        //"io/ioutil"
        "os"
        )

type Stats struct {
    simulations []*SingelRun
    name        string
}

type SingelRun struct {
    avgFitness []float64
    bestFitness []float64
    stdDevFitness []float64
    avgLearnChild []float64
    bestLearnChild []float64
    stdDevLearnChild []float64
    avgMemoryLimit []float64
    bestMemoryLimit []float64
    stdDevMemoryLimit []float64
    avgSumWeights []float64
    bestSumWeights []float64
    stdDevSumWeights []float64
    avgAge []float64
    bestAge []float64
    stdDevAge []float64
    avgLearnability []float64
    bestLearnability []float64
    stdDevLearnability []float64
    avgExtrovert []float64
    bestExtrovert []float64
    stdDevExtrovert []float64
    avgSpeech []float64
    bestSpeech []float64
    stdDevSpeech []float64
    knownUtterancesCount  []float64
    avgMemorySize  []float64
    avgMemoryWeightSize  []float64
    knownUtterances  []map[Utterance]int
    dialogsSuccessAvg  []float64
    dialogsTotalAvg  []float64
    snapshots map[int]*Snapshot
    avgLearningBiases  [][]float64
    avgDegree   []float64
    avgClusteringCoef []float64
}

func NewStats() *Stats {
    simulations := make([]*SingelRun, GlobalOptions.numberOfSimulations)
    numberOfSnapshots := int(GlobalOptions.numberOfSimulations/GlobalOptions.statsSnapshot)
    for i := 0; i < GlobalOptions.numberOfSimulations; i++ {
        simulations[i] = NewSingelRun(GlobalOptions.maxGenerations, numberOfSnapshots)
    }
    date := strconv.FormatInt(time.Now().Unix(), 10)
    // genereate folder for data
    folder := "data/"+date
    os.Mkdir(folder, 0774)
    return &Stats{simulations: simulations,
                  name: date}
}

func NewSingelRun(maxGenerations int, numberOfSnapshots int) *SingelRun {
    avgFitness := make([]float64, maxGenerations)
    bestFitness := make([]float64, maxGenerations)
    stdDevFitness := make([]float64, maxGenerations)
    avgLearnChild := make([]float64, maxGenerations)
    bestLearnChild := make([]float64, maxGenerations)
    stdDevLearnChild := make([]float64, maxGenerations)
    avgMemoryLimit := make([]float64, maxGenerations)
    bestMemoryLimit := make([]float64, maxGenerations)
    stdDevMemoryLimit := make([]float64, maxGenerations)
    avgSumWeights := make([]float64, maxGenerations)
    bestSumWeights := make([]float64, maxGenerations)
    stdDevSumWeights := make([]float64, maxGenerations)
    avgAge := make([]float64, maxGenerations)
    bestAge := make([]float64, maxGenerations)
    stdDevAge := make([]float64, maxGenerations)
    avgLearnability := make([]float64, maxGenerations)
    bestLearnability := make([]float64, maxGenerations)
    stdDevLearnability := make([]float64, maxGenerations)
    avgExtrovert := make([]float64, maxGenerations)
    bestExtrovert := make([]float64, maxGenerations)
    stdDevExtrovert := make([]float64, maxGenerations)
    avgSpeech := make([]float64, maxGenerations)
    bestSpeech := make([]float64, maxGenerations)
    stdDevSpeech := make([]float64, maxGenerations)
    knownUtterancesCount := make([]float64, maxGenerations)
    knownUtterances := make([]map[Utterance]int, maxGenerations)
    avgMemorySize := make([]float64, maxGenerations)
    avgMemoryWeightSize := make([]float64, maxGenerations)
    dialogsSuccessAvg := make([]float64, maxGenerations)
    dialogsTotalAvg := make([]float64, maxGenerations)
    snapshots := make(map[int]*Snapshot, numberOfSnapshots)
    avgLearningBiases := make([][]float64, maxGenerations)
    avgDegree := make([]float64, maxGenerations)
    avgClusteringCoef := make([]float64, maxGenerations)
    return &SingelRun{avgFitness: avgFitness,
                      bestFitness: bestFitness,
                      stdDevFitness: stdDevFitness,
                      avgSumWeights: avgSumWeights,
                      bestSumWeights: bestSumWeights,
                      stdDevSumWeights: stdDevSumWeights,
                      avgAge: avgAge,
                      bestAge: bestAge,
                      stdDevAge: stdDevAge,
                      avgLearnChild: avgLearnChild,
                      bestLearnChild: bestLearnChild,
                      stdDevLearnChild: stdDevLearnChild,
                      avgMemoryLimit: avgMemoryLimit,
                      bestMemoryLimit: bestMemoryLimit,
                      stdDevMemoryLimit: stdDevMemoryLimit,
                      avgLearnability: avgLearnability,
                      bestLearnability: bestLearnability,
                      stdDevLearnability: stdDevLearnability,
                      avgExtrovert: avgExtrovert,
                      bestExtrovert: bestExtrovert,
                      stdDevExtrovert: stdDevExtrovert,
                      avgSpeech: avgSpeech,
                      bestSpeech: bestSpeech,
                      stdDevSpeech: stdDevSpeech,
                      knownUtterancesCount: knownUtterancesCount,
                      knownUtterances: knownUtterances,
                      dialogsSuccessAvg: dialogsSuccessAvg,
                      dialogsTotalAvg: dialogsTotalAvg,
                      avgLearningBiases: avgLearningBiases,
                      avgMemorySize: avgMemorySize,
                      avgMemoryWeightSize: avgMemoryWeightSize,
                      avgDegree: avgDegree,
                      avgClusteringCoef: avgClusteringCoef,
                      snapshots: snapshots}
}

func (s *Stats) AddGenerationStats(simulationNumber int, generation int, generationStats GenerationStats) {
    s.simulations[simulationNumber].dialogsSuccessAvg[generation] = generationStats.dialogsSuccessAvg
    s.simulations[simulationNumber].dialogsTotalAvg[generation] = generationStats.dialogsTotalAvg
    s.simulations[simulationNumber].avgFitness[generation] = generationStats.fitnessAvg
    s.simulations[simulationNumber].bestFitness[generation] = generationStats.fitnessBest
    s.simulations[simulationNumber].stdDevFitness[generation] = generationStats.fitnessStdDev
    s.simulations[simulationNumber].avgSumWeights[generation] = generationStats.sumWeightsAvg
    s.simulations[simulationNumber].bestSumWeights[generation] = generationStats.sumWeightsBest
    s.simulations[simulationNumber].stdDevSumWeights[generation] = generationStats.sumWeightsStdDev
    s.simulations[simulationNumber].avgAge[generation] = generationStats.ageAvg
    s.simulations[simulationNumber].bestAge[generation] = generationStats.ageBest
    s.simulations[simulationNumber].stdDevAge[generation] = generationStats.ageStdDev
    s.simulations[simulationNumber].avgMemoryLimit[generation] = generationStats.memoryLimitAvg
    s.simulations[simulationNumber].bestMemoryLimit[generation] = generationStats.memoryLimitBest
    s.simulations[simulationNumber].stdDevMemoryLimit[generation] = generationStats.memoryLimitStdDev
    s.simulations[simulationNumber].avgLearnChild[generation] = generationStats.learnChildAvg
    s.simulations[simulationNumber].bestLearnChild[generation] = generationStats.learnChildBest
    s.simulations[simulationNumber].stdDevLearnChild[generation] = generationStats.learnChildStdDev
    s.simulations[simulationNumber].avgLearnability[generation] = generationStats.learnabilityAvg
    s.simulations[simulationNumber].bestLearnability[generation] = generationStats.learnabilityBest
    s.simulations[simulationNumber].stdDevLearnability[generation] = generationStats.learnabilityStdDev
    s.simulations[simulationNumber].avgExtrovert[generation] = generationStats.extrovertAvg
    s.simulations[simulationNumber].bestExtrovert[generation] = generationStats.extrovertBest
    s.simulations[simulationNumber].stdDevExtrovert[generation] = generationStats.extrovertStdDev
    s.simulations[simulationNumber].avgSpeech[generation] = generationStats.speechAvg
    s.simulations[simulationNumber].bestSpeech[generation] = generationStats.speechBest
    s.simulations[simulationNumber].stdDevSpeech[generation] = generationStats.speechStdDev
    s.simulations[simulationNumber].knownUtterancesCount[generation] = generationStats.knownUtterancesCount
    s.simulations[simulationNumber].knownUtterances[generation] = generationStats.knownUtterances
    s.simulations[simulationNumber].avgMemorySize[generation] = generationStats.avgMemorySize
    s.simulations[simulationNumber].avgMemoryWeightSize[generation] = generationStats.avgMemoryWeightSize
    s.simulations[simulationNumber].avgDegree[generation] = generationStats.degreeAvg
    s.simulations[simulationNumber].avgClusteringCoef[generation] = generationStats.clusteringCoefAvg
    s.simulations[simulationNumber].avgLearningBiases[generation] = generationStats.avgLearningBiases
}

func (s *Stats) PrintSummary() {
    for i, singelRun := range s.simulations {
        fmt.Printf("Simulation: %d\n", i)
        for utterance, number := range singelRun.knownUtterances[GlobalOptions.maxGenerations-1] {
            if number > 10 {
                fmt.Println(string(utterance) + ": " + IntToString(number))
            }
        }
    }
    fmt.Println("Statsfolder:" + s.name)
}

func (s *Stats) MakeSimulationSummaryGraphs() {
    for sim, _ := range s.simulations {
        //title := "TournamentSelect: " + FloatToString(GlobalOptions.turnamentSlection)
        title := "Summary " + GlobalOptions.notation
        graph := NewGraph(title, false, false)
        dialogsSuccessRatio := MakeRatioArray(s.simulations[sim].dialogsSuccessAvg,
                                              s.simulations[sim].dialogsTotalAvg)
        graph.AddX(s.simulations[sim].avgFitness, "Avg fitness")
        if(GlobalOptions.structure == "graph") {
            graph.AddX(s.simulations[sim].avgClusteringCoef, "Avg clustering coef.")
        }
        graph.AddX(dialogsSuccessRatio, "Dialogs success")
        graph.SetLabels("Generations", "Arbitrary")
        //graph.SetYRange(0,1)
        folder := "data/"+s.name
        graph.Save(folder + "/graphSimSummary" + LeadingZeros(2,sim))
    }
}

func (s *Stats) MakeGenomeSummaryGraphs() {
    for sim, _ := range s.simulations {
        //title := "TournamentSelect: " + FloatToString(GlobalOptions.turnamentSlection)
        title := "Genotypes" + GlobalOptions.notation
        graph := NewGraph(title, false, false)
        graph.AddX(s.simulations[sim].avgExtrovert, "Avg extrovert")
        if(GlobalOptions.fixedLearnability == false) {
            graph.AddX(s.simulations[sim].avgLearnability, "Avg learnability")
        }
        if(GlobalOptions.evolveSpeech == true) {
            graph.AddX(s.simulations[sim].avgSpeech, "Avg speech ability")
        }
        graph.AddX(s.simulations[sim].avgMemoryLimit, "Avg memory limit")
        graph.AddX(s.simulations[sim].avgLearnChild, "Avg teach child prob")
        graph.SetLabels("Generations", "Arbitrary")
        graph.SetYRange(0,1)
        folder := "data/"+s.name
        graph.Save(folder + "/graphGeneSummary" + LeadingZeros(2,sim))
    }
}

func (s *Stats) MakeGraphGraphs() {
    if(GlobalOptions.structure == "graph") {
        for sim, _ := range s.simulations {
            title := "Network"
            graph := NewGraph(title, false, false)
            graph.AddX(s.simulations[sim].avgDegree, "Avg degree")
            graph.SetLabels("Generations", "Degree")
            folder := "data/"+s.name
            graph.Save(folder + "/graphGraph" + LeadingZeros(2,sim))
        }
    }
}

func (s *Stats) MakeFitnessGraphs() {
    if(GlobalOptions.structure == "graph") {
         
    }
    if(GlobalOptions.structure == "lattice") {
        for sim, _ := range s.simulations {
            //title := "TournamentSelect: " + FloatToString(GlobalOptions.turnamentSlection)
            title := "Fitness"
            graph := NewGraph(title, false, false)
            graph.AddX(s.simulations[sim].bestFitness, "Best fitness")
            graph.AddX(s.simulations[sim].avgFitness, "Avg fitness")
            graph.AddX(s.simulations[sim].stdDevFitness, "Std dev fitness")
            graph.SetLabels("Generations", "Fitness")
            //graph.SetYRange(0,1)
            folder := "data/"+s.name
            graph.Save(folder + "/graphFitness" + LeadingZeros(2,sim))
        }
    }
}

func (s *Stats) MakeSumWeightsGraphs() {
    for sim, _ := range s.simulations {
        //title := "TournamentSelect: " + FloatToString(GlobalOptions.turnamentSlection)
        title := "Sum of weights"
        graph := NewGraph(title, false, false)
        graph.AddX(s.simulations[sim].bestSumWeights, "Highest sum weights")
        graph.AddX(s.simulations[sim].avgSumWeights, "Avg sum weights")
        graph.AddX(s.simulations[sim].stdDevSumWeights, "Std dev sum weights")
        graph.SetLabels("Generations", "SumWeights")
        //graph.SetYRange(0,1)
        folder := "data/"+s.name
        graph.Save(folder + "/graphSumWeights" + LeadingZeros(2,sim))
    }
}

func (s *Stats) MakeAgeGraphs() {
    for sim, _ := range s.simulations {
        //title := "TournamentSelect: " + FloatToString(GlobalOptions.turnamentSlection)
        title := "Age"
        graph := NewGraph(title, false, false)
        //graph.AddX(s.simulations[sim].bestAge, "Highest age")
        graph.AddX(s.simulations[sim].avgAge, "Avg age")
        graph.AddX(s.simulations[sim].stdDevAge, "Std dev age")
        graph.SetLabels("Generations", "Age")
        //graph.SetYRange(0,1)
        folder := "data/"+s.name
        graph.Save(folder + "/graphAge" + LeadingZeros(2,sim))
    }
}

func (s *Stats) MakeLearnChildGraphs() {
    for sim, _ := range s.simulations {
        title := "Teach child probability"
        graph := NewGraph(title, false, false)
        graph.AddX(s.simulations[sim].bestLearnChild, "Highest teach child prob")
        graph.AddX(s.simulations[sim].avgLearnChild, "Avg teach child prob")
        graph.AddX(s.simulations[sim].stdDevLearnChild, "Std dev teach child prob")
        graph.SetLabels("Generations", "Probability")
        graph.SetYRange(0,1)
        folder := "data/"+s.name
        graph.Save(folder + "/graphLearnChild" + LeadingZeros(2,sim))
    }
}

func (s *Stats) MakeMemoryLimitGraphs() {
    for sim, _ := range s.simulations {
        title := "Memory limit"
        graph := NewGraph(title, false, false)
        graph.AddX(s.simulations[sim].bestMemoryLimit, "Highest limit")
        graph.AddX(s.simulations[sim].avgMemoryLimit, "Avg limit")
        graph.AddX(s.simulations[sim].stdDevMemoryLimit, "Std dev limit")
        graph.SetLabels("Generations", "Limit * 20")
        graph.SetYRange(0,1)
        folder := "data/"+s.name
        graph.Save(folder + "/graphMemoryLimit" + LeadingZeros(2,sim))
    }
}

func (s *Stats) MakeLearnabilityGraphs() {
    for sim, _ := range s.simulations {
        //title := "TournamentSelect: " + FloatToString(GlobalOptions.turnamentSlection)
        title := "Learnability"
        graph := NewGraph(title, false, false)
        graph.AddX(s.simulations[sim].bestLearnability, "Highest learnability")
        graph.AddX(s.simulations[sim].avgLearnability, "Avg learnability")
        graph.AddX(s.simulations[sim].stdDevLearnability, "Std dev learnability")
        graph.SetLabels("Generations", "Learnability")
        //graph.SetYRange(0,1)
        folder := "data/"+s.name
        graph.Save(folder + "/graphLearnability" + LeadingZeros(2,sim))
    }
}

func (s *Stats) MakeExtrovertGraphs() {
    for sim, _ := range s.simulations {
        //title := "TournamentSelect: " + FloatToString(GlobalOptions.turnamentSlection)
        title := "Extroversion"
        graph := NewGraph(title, false, false)
        graph.AddX(s.simulations[sim].bestExtrovert, "Highest extrovert")
        graph.AddX(s.simulations[sim].avgExtrovert, "Avg extrovert")
        graph.AddX(s.simulations[sim].stdDevExtrovert, "Std dev extrovert")
        graph.SetLabels("Generations", "Extrovert")
        //graph.SetYRange(0,1)
        folder := "data/"+s.name
        graph.Save(folder + "/graphExtrovert" + LeadingZeros(2,sim))
    }
}

func (s *Stats) MakeSpeechGraphs() {
    for sim, _ := range s.simulations {
        //title := "TournamentSelect: " + FloatToString(GlobalOptions.turnamentSlection)
        title := "Speech ability"
        graph := NewGraph(title, false, false)
        graph.AddX(s.simulations[sim].bestSpeech, "Highest speech ability")
        graph.AddX(s.simulations[sim].avgSpeech, "Avg speech ability")
        graph.AddX(s.simulations[sim].stdDevSpeech, "Std dev speech ability")
        graph.SetLabels("Generations", "Speech ability")
        //graph.SetYRange(0,1)
        folder := "data/"+s.name
        graph.Save(folder + "/graphSpeech" + LeadingZeros(2,sim))
    }
}

func (s *Stats) MakeUtteranceGraphs() {
    for sim, _ := range s.simulations {
        //title := "TournamentSelect: " + FloatToString(GlobalOptions.turnamentSlection)
        title := "Utterances"
        graph := NewGraph(title, false, false)
        graph.AddX(s.simulations[sim].knownUtterancesCount, "Unique utterances")
        graph.AddX(s.simulations[sim].avgMemorySize, "Avg words in memory pr agent")
        //graph.AddX(s.simulations[sim].avgMemoryWeightSize, "Avg memory weight pr agent")
        graph.SetLabels("Generations", "Utterance")
        //graph.SetYRange(0,1)
        folder := "data/"+s.name
        graph.Save(folder + "/graphUtterance" + LeadingZeros(2,sim))
    }
}

func (s *Stats) MakeDialogGraphs() {
    for sim, _ := range s.simulations {
        //title := "TournamentSelect: " + FloatToString(GlobalOptions.turnamentSlection)
        title := "Dialogs"
        graph := NewGraph(title, false, false)
        graph.AddX(s.simulations[sim].dialogsTotalAvg, "Dialogs total avg")
        graph.AddX(s.simulations[sim].dialogsSuccessAvg, "Dialogs success avg")
        graph.SetLabels("Generations", "Dialogs")
        //graph.SetYRange(0,1)
        folder := "data/"+s.name
        graph.Save(folder + "/graphDialogs" + LeadingZeros(2,sim))
    }
}

//func (s *Stats) MakeLearningBiasesGraphs() {
//    for sim, _ := range s.simulations {
//        title := "Learning Biases"
//        graph := NewGraph(title, false, false)
//        graph.SetLabels("Generations", "Bias")
//        graph.RemoveLegend()
//        for biasNumber := 0; biasNumber < len(GlobalOptions.geneScheme)-1; biasNumber++ { // todo
//            learningBias := make([]float64, GlobalOptions.maxGenerations)
//            for i, learningBiases := range s.simulations[sim].avgLearningBiases {
//                learningBias[i] = learningBiases[biasNumber]
//            }
//            graph.AddX(learningBias, "Learning bias "+string(biasNumber))
//        }
//        //graph.SetYRange(0,1)
//        folder := "data/"+s.name
//        graph.Save(folder + "/graphLearningBiases" + LeadingZeros(2,sim))
//    }
//}

func (s *Stats) WriteFitnessStatsToFiles() {
    folder := "data/"+s.name

    // write data files
    for r, _ := range s.simulations {
        f, err := os.Create(folder + "/fitness" + strconv.Itoa(r) + ".dat")
        if err != nil {
            panic(err)
        }
        defer f.Close()
        for i, _ := range s.simulations[r].bestFitness {
            //fmt.Println(item)
            _, err := f.WriteString(
                            strconv.FormatFloat(s.simulations[r].bestFitness[i], 'f', -1, 64) + "\t" +
                            strconv.FormatFloat(s.simulations[r].avgFitness[i], 'f', -1, 64) + "\t" +
                            strconv.FormatFloat(s.simulations[r].stdDevFitness[i], 'f', -1, 64) + "\t" +
                            "\n"); 
            //file.Write([]byte(item)); 
            if err != nil {
                //fmt.Println("debug")
                fmt.Println(err)
                break
            }
        }
        // copy current files to folder 'current'
    }
    
}

func (s *Stats) AddSnapshot(sim int, snapshot *Snapshot) {
    s.simulations[sim].snapshots[snapshot.generation] = snapshot
}

func (s *Stats) PlotFitnessHeatMapForSimulation(sim int) {
    snapshots := s.simulations[sim].snapshots
    for _ , snapshot := range snapshots {
        title := "Fitness"
        folder := "data/"+s.name+"/"
        filename := "fitnessHeatMapSim" + 
                     IntToString(sim) + "_" + LeadingZeros(6, snapshot.generation)
        if(GlobalOptions.structure == "graph") {
            g := gographviz.NewGraph()
            g.SetName("G")
            g.SetDir(false)
            for _, agent := range snapshot.agentsAlive {
                a := gographviz.NewAttrs()
                a.Add("shape", "dot")
                a.Add("value", FloatToString(0.1+agent.GetFitness())) 
                // 0.1 so poor fitted agents show up at all
                a.Add("group", string(agent.GetHighestWeightedUtterance()))
                g.AddNode("G", IntToString(agent.GetId()), a)
                //fmt.Println(IntToString(agent.GetId()), string(agent.GetHighestWeightedUtterance()))
            }
            for agentPair, weight := range snapshot.relations {
                a := gographviz.NewAttrs()
                a.Add("label", FloatToString(weight))
                g.AddEdge(IntToString(agentPair.a.GetId()), "", IntToString(agentPair.b.GetId()), "", false, a)
            }
            s := g.String()
            ioutil.WriteFile(folder+filename+".dot", []byte(s), 0666)
        }
        if(GlobalOptions.structure == "lattice") {
                agents := snapshot.lattice
                size := float64(len(agents))
                // save every snapshot
                fitnessHeatMap := make([][]float64, len(agents))
                for x := range agents {
                    fitnessHeatMap[x] = make([]float64, len(agents))
                    for y := range agents[x] {
                        fitnessHeatMap[x][y] = agents[x][y].GetFitness()
                    }
                }
                heatMap := NewHeatMap(title, false, false)
                heatMap.Save(filename)
                heatMap.SetRanges(size, size, 1, "Fitness")
                heatMap.UpdateMap(fitnessHeatMap)
                heatMap.Close()
                Move(filename+".png", folder)
            }
        }
}

func (s *Stats) PlotHighestWeightedUtteranceHeatMapForSimulation(sim int) {
    snapshots := s.simulations[sim].snapshots
    for _ , snapshot := range snapshots {
        if(GlobalOptions.structure == "lattice") {
            agents := snapshot.lattice
            size := float64(len(agents))
            // save every snapshot
            fitnessHeatMap := make([][]float64, len(agents))
            for x := range agents {
                fitnessHeatMap[x] = make([]float64, len(agents))
                for y := range agents[x] {
                    fitnessHeatMap[x][y] = 1
                    _ = agents[x][y].GetHighestWeightedUtterance()
                }
            }
            title := "Highest weighted utterances"
            folder := "data/"+s.name+"/"
            filename := "highestWeightedUtteranceHeatMapSim" + 
                         IntToString(sim) + "_" + LeadingZeros(6, snapshot.generation)
            heatMap := NewHeatMap(title, false, false)
            heatMap.Save(filename)
            //genomeLength := len(GlobalOptions.geneScheme)
            //maxIntegerFromGenesLength := math.Pow(2.0, float64(GlobalOptions.populationSize))
            heatMap.SetRanges(size, size, float64(GlobalOptions.populationSize), "Arbitrary")
            heatMap.UpdateMap(fitnessHeatMap)
            heatMap.Close()
            Move(filename+".png", folder)
        }
    }
}


func (s *Stats) PlotLearnabilityRateHeatMapForSimulation(sim int) {
    snapshots := s.simulations[sim].snapshots
    for _ , snapshot := range snapshots {
        if(GlobalOptions.structure == "lattice") {
            agents := snapshot.lattice
            size := float64(len(agents))
            // save every snapshot
            fitnessHeatMap := make([][]float64, len(agents))
            for x := range agents {
                fitnessHeatMap[x] = make([]float64, len(agents))
                for y := range agents[x] {
                    learnabiliyRate := agents[x][y].GetLearnability()
                    fitnessHeatMap[x][y] = learnabiliyRate
                }
            }
            title := "Learnability rate, generation "+IntToString(snapshot.generation)
            folder := "data/"+s.name+"/"
            filename := "LearnabilityRateHeatMapSim" + 
                         IntToString(sim) + "_" + LeadingZeros(6, snapshot.generation)
            heatMap := NewHeatMap(title, false, false)
            heatMap.Save(filename)
            heatMap.SetRanges(size, size, 2, "Learnability pr agent")
            heatMap.UpdateMap(fitnessHeatMap)
            heatMap.Close()
            Move(filename+".png", folder)
        }
    }
}

func (s *Stats) CopyStatsFolderToCurrent() {
    if err := exec.Command("/bin/cp", "-r", "http/statstemplate/index.html", "data/"+s.name+"/").Run(); err != nil {
        Log("Could not copy data to current")
    }
    folder := "data/"+s.name+"/"
    if err := exec.Command("/bin/rm", "-r", "data/current").Run(); err != nil {
        Log("Could not delete data/current")
    }
    if err := exec.Command("/bin/cp", "-r", folder, "data/current").Run(); err != nil {
        Log("Could not copy data to current")
    }
}

func (s *Stats) GenerateStatsFromGeneration(agentList []*Agent) GenerationStats {
    populationSize = len(agentList)
    dialogSuccessAccum := 0
    dialogTotalAccum := 0
    fitnessAvg := 0.0
    fitnessBest := 0.0
    learnabilityAvg := 0.0
    learnabilityBest := 0.0
    costAvg := 0.0
    costBest := 0.0
    extrovertAvg := 0.0
    extrovertBest := 0.0
    speechAvg := 0.0
    speechBest := 0.0
    ageAvg := 0.0
    ageBest := 0.0
    learnChildAvg := 0.0
    learnChildBest := 0.0
    memoryLimitAvg := 0.0
    memoryLimitBest := 0.0
    sumWeightsAvg := 0.0
    sumWeightsBest := 0.0
    //learningBiases := make([][]float64, populationSize)
    // Get avg and best numbers
    counter := 0
    for i := range agentList {
        if agentList[i].alive == false {
            msg := "Error: Trying to find stats on population where there are dead agents"
            Log(msg)
            fmt.Println(msg)
        }
        agent := agentList[i]
        dialogSuccessAccum += agent.GetDialogSuccess()
        dialogTotalAccum += agent.GetDialogTotal()
        memoryLimitAvg += agent.GetMemoryLimit()
        learnChildAvg += agent.GetLearnChildProbability()
        fitnessAvg += agent.GetFitness()
        sumWeightsAvg += agent.GetSumWeights()
        learnabilityAvg += agent.GetLearnability()
        extrovertAvg += agent.SpeakToForeignerChance()
        speechAvg += agent.GetSpeechAbility()
        ageAvg += float64(agent.GetAge())
        if (agent.GetFitness() > fitnessBest) {
            fitnessBest = agent.GetFitness()
        }
        if (agent.GetMemoryLimit() > memoryLimitBest) {
            memoryLimitBest = agent.GetMemoryLimit()
        }
        if (agent.GetLearnChildProbability() > learnChildBest) {
            learnChildBest = agent.GetLearnChildProbability()
        }
        if (agent.GetSumWeights() > sumWeightsBest) {
            sumWeightsBest = agent.GetSumWeights()
        }
        if (float64(agent.GetAge()) > ageBest) {
            ageBest = float64(agent.GetAge())
        }
        if (agent.GetLearnability() > learnabilityBest) {
            learnabilityBest = agent.GetLearnability()
        }
        //learningBiases[counter] = agent.GetLearningBiases()
        counter++
        // log evry agent
        //pl.agentLog[agent.GetId()] += agent.String()
    }
    dialogSuccessAvg := float64(dialogSuccessAccum)/float64(populationSize)
    dialogTotalAvg := float64(dialogTotalAccum)/float64(populationSize)
    memoryLimitAvg = memoryLimitAvg/float64(populationSize)
    learnChildAvg = learnChildAvg/float64(populationSize)
    fitnessAvg = fitnessAvg/float64(populationSize)
    sumWeightsAvg = sumWeightsAvg/float64(populationSize)
    // hack
    GlobalOptions.averageLexiconWeights = sumWeightsAvg
    ageAvg = float64(ageAvg)/float64(populationSize)
    learnabilityAvg = learnabilityAvg/float64(populationSize)
    //avgLearningBiases := AvgArray(learningBiases)
    costAvg = costAvg/float64(populationSize)
    extrovertAvg = extrovertAvg/float64(populationSize)
    speechAvg = speechAvg/float64(populationSize)

    fitnessStdDev := 0.0
    memoryLimitStdDev := 0.0
    learnChildStdDev := 0.0
    sumWeightsStdDev := 0.0
    learnabilityStdDev := 0.0
    extrovertStdDev := 0.0
    speechStdDev := 0.0
    costStdDev := 0.0
    ageStdDev := 0.0
    // Get standard dev numbers, req pre avg cac
    for _, agent := range agentList {
        learnChildStdDev       +=  math.Pow((agent.GetLearnChildProbability()       -  learnChildAvg),       2)
        memoryLimitStdDev       +=  math.Pow((agent.GetMemoryLimit()       -  memoryLimitAvg),       2)
        fitnessStdDev       +=  math.Pow((agent.GetFitness()       -  fitnessAvg),       2)
        sumWeightsStdDev    +=  math.Pow((agent.GetSumWeights()    -  sumWeightsAvg),    2)
        learnabilityStdDev  +=  math.Pow((agent.GetLearnability()  -  learnabilityAvg),  2)
        extrovertStdDev     +=  math.Pow((agent.SpeakToForeignerChance()          -  extrovertAvg),          2)
        speechStdDev     +=  math.Pow((agent.GetSpeechAbility()          -  speechAvg),          2)
        ageStdDev           +=  math.Pow((float64(agent.GetAge())  -  ageAvg),           2)
    }
    memoryLimitStdDev = math.Sqrt(1.0/float64(populationSize) * memoryLimitStdDev)
    learnChildStdDev = math.Sqrt(1.0/float64(populationSize) * learnChildStdDev)
    fitnessStdDev = math.Sqrt(1.0/float64(populationSize) * fitnessStdDev)
    sumWeightsStdDev = math.Sqrt(1.0/float64(populationSize) * sumWeightsStdDev)
    ageStdDev = math.Sqrt(1.0/float64(populationSize) * ageStdDev)
    learnabilityStdDev = math.Sqrt(1.0/float64(populationSize) * learnabilityStdDev)
    costStdDev = math.Sqrt(1.0/float64(populationSize) * costStdDev)
    extrovertStdDev = math.Sqrt(1.0/float64(populationSize) * extrovertStdDev)
    speechStdDev = math.Sqrt(1.0/float64(populationSize) * speechStdDev)

    // put in over
    knownUtterances := make(map[Utterance]int)
    avgMemorySize := 0.0
    avgMemoryWeightSize := 0.0
    //sumWeightUtterance := make(map[Utterance]float64)
    for _,agent := range agentList {
       // memoryPointer := pl.lattice[x][y].GetMemory() 
       // for utterance, weight := range *memoryPointer {
       //     knownUtterances[utterance] += 1
       //     sumWeightUtterance[utterance] += weight
       // }
       knownUtterances[agent.GetHighestWeightedUtterance()] += 1
       avgMemorySize += float64(agent.GetNumberOfKnownUtterances())
       avgMemoryWeightSize += float64(agent.GetSumWeights())
    }
    avgMemorySize = avgMemorySize / float64(len(agentList))
    avgMemoryWeightSize = avgMemorySize / float64(len(agentList))

    // return sturct
    return GenerationStats{
        dialogsSuccessAvg: dialogSuccessAvg,
        dialogsTotalAvg: dialogTotalAvg,
        fitnessAvg: fitnessAvg,
        fitnessStdDev: fitnessStdDev,
        fitnessBest: fitnessBest,
        memoryLimitAvg: memoryLimitAvg,
        memoryLimitStdDev: memoryLimitStdDev,
        memoryLimitBest: memoryLimitBest,
        learnChildAvg: learnChildAvg,
        learnChildStdDev: learnChildStdDev,
        learnChildBest: learnChildBest,
        sumWeightsAvg: sumWeightsAvg,
        sumWeightsStdDev: sumWeightsStdDev,
        sumWeightsBest: sumWeightsBest,
        ageAvg: ageAvg,
        ageStdDev: ageStdDev,
        ageBest: ageBest,
        learnabilityAvg: learnabilityAvg,
        learnabilityStdDev: learnabilityStdDev,
        learnabilityBest: learnabilityBest,
        costAvg: costAvg,
        costStdDev: costStdDev,
        costBest: costBest,
        extrovertAvg: extrovertAvg,
        extrovertStdDev: extrovertStdDev,
        extrovertBest: extrovertBest,
        speechAvg: speechAvg,
        speechStdDev: speechStdDev,
        speechBest: speechBest,
        knownUtterancesCount: float64(len(knownUtterances)),
        knownUtterances: knownUtterances,
        avgMemorySize: avgMemorySize,
        avgMemoryWeightSize: avgMemoryWeightSize,
        //avgLearningBiases: avgLearningBiases,
    }
}
