package main

import "github.com/sbinet/go-gnuplot"
import "fmt"
import "bytes"
import "strconv"


func plotGraph(data map[string]([]float64)) {
    fname := ""
    persist := true
    debug := false

    plotter,err := gnuplot.NewPlotter(fname, persist, debug)
    if err != nil {
        err_string := fmt.Sprintf("** err: %v\n", err)
        panic(err_string)
    }
    for name, points := range data {
        plotter.PlotX(points, name)
    }
    plotter.SetStyle("lines")
    plotter.CheckedCmd("replot")
    defer plotter.Close()

}

func plotGraphXY(x, y []float64) {
    fname := ""
    persist := true
    debug := false

    plotter,err := gnuplot.NewPlotter(fname, persist, debug)
    if err != nil {
        err_string := fmt.Sprintf("** err: %v\n", err)
        panic(err_string)
    }
    plotter.PlotXY(x, y, "Lifetime over learnability")
    plotter.CheckedCmd("replot")
    defer plotter.Close()

}

func getNewMap() *gnuplot.Plotter {
    fname := ""
    persist := true
    debug := false

    plotter, err := gnuplot.NewPlotter(fname, persist, debug)
    if err != nil {
        err_string := fmt.Sprintf("** err: %v\n", err)
        panic(err_string)
    }
    return plotter
}

func updateLexiconMap(matrix [][]Agent, plotter *gnuplot.Plotter, globalLexicon map[string]int) {
    plotter.ResetPlot()
    plotter.Cmd("set title \"Highest weightet word pr agent\"")
    plotter.Cmd("plot '-' matrix with image\n")
    var buffer bytes.Buffer
    for x := range matrix {
        for y := range matrix[x] {
            buffer.WriteString(strconv.Itoa(globalLexicon[getWordWithHighestWeight(matrix[x][y].Lexicon)]))
            buffer.WriteString(" ")
        }
        buffer.WriteString("\n")
    }
    buffer.WriteString("e\ne\n")
    plotter.Cmd(buffer.String())
}

func updateLearnabilityMap(matrix [][]Agent, plotter *gnuplot.Plotter) {
    plotter.ResetPlot()
    plotter.Cmd("set title \"Learnability pr agent\"")
    plotter.Cmd("set cbrange [0:1]\n")
    plotter.Cmd("plot '-' matrix with image\n") 
    var buffer bytes.Buffer
    for x := range matrix {
        for y := range matrix[x] {
            if (!matrix[x][y].Alive) { buffer.WriteString("0 "); continue }
            buffer.WriteString(strconv.FormatFloat(matrix[x][y].Learnability, 'f', 6, 64))
            buffer.WriteString(" ")
        }
        buffer.WriteString("\n")
    }
    buffer.WriteString("e\ne\n")
    plotter.Cmd(buffer.String())
}

func updateSumWeightsMap(matrix [][]Agent, plotter *gnuplot.Plotter) {
    plotter.ResetPlot()
    plotter.Cmd("set title \"Sum of weights on lexicon pr agent\"")
    plotter.Cmd("set cbrange [0:20]\n")
    plotter.Cmd("plot '-' matrix with image\n") 
    var buffer bytes.Buffer
    for x := range matrix {
        for y := range matrix[x] {
            if (!matrix[x][y].Alive) { buffer.WriteString("0 "); continue }
            sumWeights := 0.0
            for _, weight := range matrix[x][y].Lexicon {
                sumWeights += weight
            }
            buffer.WriteString(strconv.FormatFloat(sumWeights, 'f', 6, 64))
            buffer.WriteString(" ")
        }
        buffer.WriteString("\n")
    }
    buffer.WriteString("e\ne\n")
    plotter.Cmd(buffer.String())
}

func updateAgeMap(matrix [][]Agent, plotter *gnuplot.Plotter) {
    plotter.ResetPlot()
    plotter.Cmd("set title \"Age pr agent\"")
    plotter.Cmd("plot '-' matrix with image\n") 
    var buffer bytes.Buffer
    for x := range matrix {
        for y := range matrix[x] {
            if (!matrix[x][y].Alive) { buffer.WriteString("0 "); continue }
            buffer.WriteString(strconv.Itoa(matrix[x][y].Age))
            buffer.WriteString(" ")
        }
        buffer.WriteString("\n")
    }
    buffer.WriteString("e\ne\n")
    plotter.Cmd(buffer.String())
}

func updateLexiconSizeMap(matrix [][]Agent, plotter *gnuplot.Plotter) {
    plotter.ResetPlot()
    plotter.Cmd("set title \"Lexicon size pr agent\"")
    plotter.Cmd("plot '-' matrix with image\n") 
    var buffer bytes.Buffer
    for x := range matrix {
        for y := range matrix[x] {
            if (!matrix[x][y].Alive) { buffer.WriteString("0 "); continue }
            buffer.WriteString(strconv.Itoa(len(matrix[x][y].Lexicon)))
            buffer.WriteString(" ")
        }
        buffer.WriteString("\n")
    }
    buffer.WriteString("e\ne\n")
    plotter.Cmd(buffer.String())
}
